library image_tool;

import 'dart:io';
import 'dart:core';
import 'package:path/path.dart' as p;
import 'package:watcher/watcher.dart';
import "package:yaml/yaml.dart";


import 'src/config.dart';

class FlutterImgResolutionOrganizer {
  /// A configulation instance
  Config config;

  /// A main constructor with a default configuration
  FlutterImgResolutionOrganizer() {
    config = Config();
  }

  /// A named constructor that creates an instance with a config
  FlutterImgResolutionOrganizer.withConfig(
      {String assetFolderPath = DefaultConfig.ASSET_FOLDER_PATH,
      List<String> fileExtensions = DefaultConfig.FILE_EXTENSIONS,
      String resolutionIndicator = DefaultConfig.RESOLUTION_INDICATOR,
      bool allowOverwrite = DefaultConfig.ALLOW_OVERWRITE}) {
    config = Config(
      assetFolderPath: assetFolderPath,
      fileExtensions: fileExtensions,
      resolutionIndicator: resolutionIndicator,
      allowOverwrite: allowOverwrite,
    );
  }

  /// A named constructor that creates an instance with a configPath provided
  FlutterImgResolutionOrganizer.withConfigFile(String configPath) {
    config = Config.fromFile(configPath);
  }

  /// [For CLI] Organise image's file according to the setting in the config instance
  Future<int> run() async {
    int exitCode = 0;

    List<FileSystemEntity> entities = _getFileEntitiesInDirectory(
        config.assetFolderPath, config.fileExtensions);

    List<String> assetsPaths = entities.map((f)=>f.path).toList();


    File file = new File(config.pubspecPath);
    List<String> addPaths = [];

    if (file?.existsSync() == true) {

      String fileStr = file.readAsStringSync();

      Map yamlMap = loadYaml(fileStr);
      YamlMap flutterNode = yamlMap['flutter'];
      YamlList assetsNode = null;
      int flutterIndex = 0;
      int assetsIndex = 0;
      if(flutterNode == null){
        //没有flutter node
        fileStr += "\n\nflutter:\n\n  assets: ";
      }else{
        assetsNode = flutterNode['assets'];
      }

      String front = null;
      bool match = false;
      while(!match){
        flutterIndex = fileStr.indexOf('flutter:',flutterIndex);
        front = fileStr.substring(flutterIndex-1,flutterIndex);

        if(front == "\n" ){
          match = true;
        }else{
          flutterIndex += 8;
        }
      }

      assetsIndex = fileStr.indexOf('assets:',flutterIndex);
      if(assetsIndex == -1){
        fileStr = fileStr.replaceRange(flutterIndex+8,flutterIndex+8,"\n  assets: ");
        assetsIndex = fileStr.indexOf('assets:',flutterIndex);
      }

      if(assetsNode == null){
        addPaths = assetsPaths;
      }else{
        bool contains = false;
        for(String path in assetsPaths){

          contains = false;
          for(String item in assetsNode.value){

            if(path == item) contains = true;
          }
          if(!contains){
            addPaths.add(path);
          }
        }
      }

      String pathStr = "";
      for(String path in addPaths){
        pathStr += "\n    - ${path}";
      }

      fileStr = fileStr.replaceRange(assetsIndex+7,assetsIndex+7,pathStr);
      await file.writeAsString(fileStr);

    }

    int nAddedFile = addPaths.length;
    print('$nAddedFile image file(s) have been added to pubspec.');

    return exitCode;
  }

  int group() {

    int exitCode = 0;

    int nMovedFile = 0;
    List<FileSystemEntity> entities = _getFileEntitiesInDirectory(
        config.assetFolderPath, config.fileExtensions);

    for (var entity in entities) {
      nMovedFile += moveFileToItsResolutionFolder(entity);
    }
    print('$nMovedFile file(s) have been arranged.');

    return exitCode;
  }

    /// [For CLI] Watch and organise image's file according to the setting in the config instance
  Future<int> watch() async {
    int exitCode = 0;

    run();
    print('Watching...');

    var watcher = DirectoryWatcher(p.absolute(config.assetFolderPath));
    await for (var event in watcher.events) {
      if (event.type == ChangeType.ADD || event.type == ChangeType.MODIFY) {
        var f = File(event.path);
        if (_matchWithFileExtensions(f)) {
          moveFileToItsResolutionFolder(f);
        }
      }
    }

    return exitCode;
  }

  /// Move a file to the target folder according to its resolution indicator
  int moveFileToItsResolutionFolder(FileSystemEntity entity) {
    var fileNameRegExp = config.fileNameRegExp;
    int nAffectedFile = 0;
    Match fileNameMatch = fileNameRegExp.firstMatch(p.basename(entity.path));
    if (fileNameMatch != null) {
      // Extract a filename and its resolution
      String fileNameWithExt =
          '${fileNameMatch.group(1)}${fileNameMatch.group(4)}${fileNameMatch.group(5)}';
      double resolution = double.parse('${fileNameMatch.group(3)}');

      String resolutionDirectoryName = '${resolution.toStringAsFixed(1)}x';
      String toDirectory = '${entity.parent.path}/$resolutionDirectoryName/';
      Directory(toDirectory).createSync(recursive: true);
      String toPath = '$toDirectory$fileNameWithExt';

      if (config.allowOverwrite ||
          FileSystemEntity.typeSync(toPath) == FileSystemEntityType.notFound) {
        entity.renameSync('$toDirectory$fileNameWithExt');
        print('Moved ${entity.path} -> $toDirectory$fileNameWithExt');
        nAffectedFile = 1;
      } else {
        print(
            '$toDirectory$fileNameWithExt has already existed. Try changing the file name.');
      }
    }
    return nAffectedFile;
  }

  /// Return a list of all files in the specific directory filtered by file extensions
  List<FileSystemEntity> _getFileEntitiesInDirectory(
      String directoryPath, List<String> fileExtensions) {
    var dir = Directory(directoryPath);
    List entities = dir
        .listSync(recursive: true, followLinks: false)
        .where((el) =>
            el is FileSystemEntity &&
            el is File &&
            _matchWithFileExtensions(el))
        .toList();
    return entities;
  }

  /// Return true if this file should be organised (match with file extensions listed in the configulation)
  bool _matchWithFileExtensions(File file) {
    var fileExtensionsSet = Set.from(config.fileExtensions);
    return fileExtensionsSet.contains(p.extension(file.path));
  }
}
